/*
 *
 * Copyright (c) 2016,2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Adrian Szyndela <adrian.s@samsung.com>
 *          Rafal Pietruch <r.pietruch@samsung.com>
 *          Łukasz Stelmach <l.stelmach@samsung.com>
 */
/**
 * @file crash-stack-arm.c
 * @brief unwinding call stacks, functions specific for ARM
 */
#include "crash-stack.h"
#include "shared/log.h"
#include "shared/util.h"

#include <elf.h>
#include <stdio.h>
#include <string.h>
#include <sys/user.h>

// definitions copied from previously used McTernan unwinder
#define REGS_REGULAR_NUM 13
#define REG_FP 11
#define REG_IP REG_ARM_PC
#define REG_SP REG_ARM_SP
#define REG_LR 14
#define REG_PC 15
#define REG_SPSR 16

/**
 * @brief Important registers for unwinding stack on ARM
 */
struct Regs {
	Elf64_Addr regs[REGS_REGULAR_NUM];		///< regular registers
	Elf64_Addr sp;							///< register: stack pointer
	Elf64_Addr lr;							///< register: link register
	Elf64_Addr pc;							///< register: program counter
	Elf64_Addr spsr;						///< register: status register
};

typedef struct Regs Regs;					///< convenience type definition
static Regs g_regs;							///< static storage for target register values

static struct user_regs g_registers;	///< static storage for registers

void *_crash_stack_get_memory_for_registers(size_t *size)
{
	if (NULL != size)
		*size = sizeof(g_registers);
	return &g_registers;
}

void *_get_place_for_register_value(const char *regname, int regnum)
{
	if (strcmp(regname, "pc") == 0 || REG_PC == regnum)
		return &g_regs.pc;
	else if (strcmp(regname, "sp") == 0 || REG_SP == regnum)
		return &g_regs.sp;
	else if (strcmp(regname, "lr") == 0 || REG_LR == regnum)
		return &g_regs.lr;
	else if (strcmp(regname, "spsr") == 0 || REG_SPSR == regnum)
		return &g_regs.spsr;
	else if (regnum < REGS_REGULAR_NUM)
		return &g_regs.regs[regnum];
	return NULL;
}

void _crash_stack_set_registers(void *regbuf)
{
	struct user_regs *registers = regbuf;
	int i;
	for (i = 0; i < sizeof(registers->uregs)/sizeof(registers->uregs[0]); i++) {
		void *regmem = _get_place_for_register_value("", i);
		if (NULL != regmem)
			memcpy(regmem, &registers->uregs[i], sizeof(registers->uregs[i]));
	}
}

unsigned long _get_register_value(int n)
{
	return g_registers.uregs[n];
}

void _crash_stack_print_regs(json_object *jobj)
{
	static struct reg {
		const char *name;
		int index;
	} regs[] = {{"r0", 0},
	            {"r1", 1},
	            {"r2", 2},
	            {"r3", 3},
	            {"r4", 4},
	            {"r5", 5},
	            {"r6", 6},
	            {"r7", 7},
	            {"r8", 8},
	            {"r9", 9},
	            {"r10", 10},
	            {"fp", REG_FP},
	            {"ip", REG_IP},
	            {"sp", REG_SP},
	            {"lr", REG_LR},
	            {"pc", REG_PC},
	            {"cpsr", REG_SPSR}}; // [> XXX the label should probably write "spsr" <]

	json_object *j_regs = json_object_new_array();

	for (size_t i = 0; i < ARRAY_SIZE(regs); i++) {
		char value_str[8+3];
		if (sprintf(value_str, "0x%08lx", g_registers.uregs[regs[i].index]) == -1) {
			_E("Cannot print register value: %m");
			continue;
		}
		ADD_REG_TO_JSON_ARRAY(j_regs, regs[i].name, value_str);
	}

	json_object_object_add(jobj, K_REGISTERS, j_regs);
}
