/*
 * Copyright (c) 2020 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <assert.h>
#include <json-c/json.h>
#include <signal.h>
#include <stdio.h>

#include "crash-stack.h"
#include "report_info.h"
#include "shared/log.h"
#include "shared/util.h"

static void print_exe_path(json_object *j_root, FILE *fp)
{
	assert(j_root);
	assert(fp);

	json_object *j_exe_path;
	if (!json_object_object_get_ex(j_root, "exe_file_path", &j_exe_path)) {
		_W("No exe file path in the JSON object");
		return;
	}

	json_object_get_string(j_exe_path);
	fprintf(fp,
		"Executable File Path: %s\n",
		json_object_get_string(j_exe_path));
}

static void print_signal(json_object *j_root, FILE *fp)
{
	assert(j_root);
	assert(fp);

#define DEF_STR(name) [name] = #name
	const char* const signal_table[] = {
		DEF_STR(SIGHUP),
		DEF_STR(SIGINT),
		DEF_STR(SIGQUIT),
		DEF_STR(SIGILL),
		DEF_STR(SIGTRAP),
		DEF_STR(SIGABRT),
		DEF_STR(SIGBUS),
		DEF_STR(SIGFPE),
		DEF_STR(SIGKILL),
		DEF_STR(SIGUSR1),
		DEF_STR(SIGUSR2),
		DEF_STR(SIGPIPE),
		DEF_STR(SIGSEGV),
		DEF_STR(SIGTERM),
		DEF_STR(SIGSTKFLT),
		DEF_STR(SIGCHLD),
		DEF_STR(SIGCONT),
		DEF_STR(SIGSTOP),
		DEF_STR(SIGTSTP),
		DEF_STR(SIGTTIN),
		DEF_STR(SIGTTOU),
		DEF_STR(SIGURG),
		DEF_STR(SIGXCPU),
		DEF_STR(SIGXFSZ),
		DEF_STR(SIGVTALRM),
		DEF_STR(SIGPROF),
		DEF_STR(SIGWINCH),
		DEF_STR(SIGPOLL),
		DEF_STR(SIGIO),
		DEF_STR(SIGPWR),
		DEF_STR(SIGSYS)
	};
#undef DEF_STR
	json_object *j_signal;
	if (!json_object_object_get_ex(j_root, K_SIGNAL, &j_signal)) {
		_W("No signal info in the JSON object");
		return;
	}
	int signo = json_object_get_int(j_signal);
	if (signo < 0 || signo >= ARRAY_SIZE(signal_table)) {
		_E("Invalid signal number (%d) in the JSON object", signo);
		return;
	}
	fprintf(fp,
		"Signal: %d\n"
		"      (%s)\n",
		signo,
		signal_table[signo]);
}

static void print_registers(json_object *j_root, FILE *fp)
{
	assert(j_root);
	assert(fp);

	json_object *j_regs;
	if (!json_object_object_get_ex(j_root, K_REGISTERS, &j_regs)) {
		_W("No registers info in the JSON object");
		return;
	}

	int len = json_object_array_length(j_regs);
	fprintf(fp, "\nRegister Information\n");
	for (int i = 0; i < len; i++) {
		json_object *j_entry = json_object_array_get_idx(j_regs, i);
		json_object *j_name, *j_value;
		if (!json_object_object_get_ex(j_entry, K_REGS_NAME, &j_name) ||
		    !json_object_object_get_ex(j_entry, K_REGS_VALUE, &j_value)) {
			continue;
		}

		fprintf(fp,
		        "%-4s = %s%s",
		        json_object_get_string(j_name),
		        json_object_get_string(j_value),
		        ((i + 1) % 2 == 0 || i == len - 1) ? "\n" : ", ");
	}
}

static void print_memory_info(json_object *j_root, FILE *fp)
{
	assert(j_root);
	assert(fp);

	fprintf(fp, "\nMemory information\n");
	json_object *j_mem;

	if (!json_object_object_get_ex(j_root, K_MEM_INFO, &j_mem)) {
		_W("No memory info in the JSON object");
		return;
	}

	static struct infoname {
		const char *name, *keyname;
	} names[] = {{"MemTotal:", K_MI_MEMTOTAL},
	             {"MemFree:",  K_MI_MEMFREE},
	             {"Buffers:",  K_MI_BUFFS},
	             {"Cached:",   K_MI_CACHED},
	             {"VmPeak:",   K_MI_VMPEAK},
	             {"VmSize:",   K_MI_VMSIZE},
	             {"VmLck:",    K_MI_VMLCK},
	             {"VmPin:",    K_MI_VMPIN},
	             {"VmHWM:",    K_MI_VMHWM},
	             {"VmRSS:",    K_MI_VMRSS},
	             {"VmData:",   K_MI_VMDATA},
	             {"VmStk:",    K_MI_VMSTK},
	             {"VmExe:",    K_MI_VMEXE},
	             {"VmLib:",    K_MI_VMLIB},
	             {"VmPTE:",    K_MI_VMPTE},
	             {"VmSwap:",   K_MI_VMSWAP}};

	static const char info_len = 8 + 1; // longest info description

	for (size_t i = 0; i < ARRAY_SIZE(names); i++) {
		json_object *j_info;
		if (!json_object_object_get_ex(j_mem, names[i].keyname, &j_info)) {
			_W("No %s in the JSON object", names[i].keyname);
			continue;
		}

		fprintf(fp,
			"%-*s %8s KB\n",
			info_len,
			names[i].name,
			json_object_get_string(j_info));
	}
}

static void print_threads(json_object *j_root, FILE *fp)
{
	assert(j_root);
	assert(fp);

	json_object *j_threads, *j_tasks;
	if (!json_object_object_get_ex(j_root, K_THREADS, &j_threads) ||
	    !json_object_object_get_ex(j_threads, K_TH_TASKS, &j_tasks)) {
		_W("No thread info in the JSON object");
		return;
	}

	int tasks_len = json_object_array_length(j_tasks);
	if (tasks_len <= 1)
		return;

	json_object *j_pid, *j_tid;
	if (!json_object_object_get_ex(j_threads, K_TH_PID, &j_pid) ||
	    !json_object_object_get_ex(j_threads, K_TH_TID, &j_tid)) {
		_W("No PID or TID info in the JSON object");
		return;
	}

	fprintf(fp,
	        "\nThreads Information\n"
	        "Threads: %d\n"
	        "PID = %d TID = %d\n",
	        tasks_len,
	        json_object_get_int(j_pid),
	        json_object_get_int(j_tid));

	for (int i = 0; i < tasks_len; i++) {
		json_object *j_task_id = json_object_array_get_idx(j_tasks, i);
		fprintf(fp,
		        "%d%s",
		        json_object_get_int(j_task_id),
		        i == tasks_len - 1 ? "\n" : " ");
	}
}

static void print_maps(json_object *j_root, FILE *fp)
{
	assert(j_root);
	assert(fp);

	json_object *j_maps;
	if (!json_object_object_get_ex(j_root, K_MAPS, &j_maps)) {
		_W("No maps info in the JSON object");
		return;
	}

	fprintf(fp, "\nMaps Information\n");
	int maps_len = json_object_array_length(j_maps);
	for (int i = 0; i < maps_len; i++) {
		json_object *j_entry = json_object_array_get_idx(j_maps, i);
		json_object *j_start, *j_end, *j_perm, *j_path;
		if (!json_object_object_get_ex(j_entry, K_MAPS_START, &j_start) ||
		    !json_object_object_get_ex(j_entry, K_MAPS_END, &j_end) ||
		    !json_object_object_get_ex(j_entry, K_MAPS_PERM, &j_perm) ||
		    !json_object_object_get_ex(j_entry, K_MAPS_PATH, &j_path))
			continue;

		unsigned long long start, end;
		sscanf(json_object_get_string(j_start), "%llx", &start);
		sscanf(json_object_get_string(j_end), "%llx", &end);
		fprintf(fp,
		        "%16llx %16llx %s %s\n",
		        start,
		        end,
		        json_object_get_string(j_perm),
		        json_object_get_string(j_path));
	}
	fprintf(fp, "End of Maps Information\n");
}

static void print_callstack(json_object *j_root, FILE *fp)
{
	assert(j_root);
	assert(fp);

	json_object *j_callstack, *j_threads, *j_pid;
	if (!json_object_object_get_ex(j_root, K_CALLSTACK, &j_callstack) ||
	    !json_object_object_get_ex(j_root, K_THREADS, &j_threads) ||
	    !json_object_object_get_ex(j_threads, K_TH_PID, &j_pid)) {
		printf("No callstack info in the JSON object");
		return;
	}

	int callstack_len = json_object_array_length(j_callstack);

	fprintf(fp, "\nCallstack Information (PID:%d)\n"
	            "Call Stack Count: %d\n",
	            json_object_get_int(j_pid),
	            callstack_len);

	for (int i = 0; i < callstack_len; i++) {
		json_object *j_entry = json_object_array_get_idx(j_callstack, i);
		json_object *j_name, *j_addr, *j_m_name, *j_m_offset, *j_offset;

		if (!json_object_object_get_ex(j_entry, K_CALL_ADDR, &j_addr) ||
		    !json_object_object_get_ex(j_entry, K_CALL_M_NAME, &j_m_name) ||
		    !json_object_object_get_ex(j_entry, K_CALL_M_OFFSET, &j_m_offset))
			continue;

		fprintf(fp, "%2d: ", i);
		if (json_object_object_get_ex(j_entry, K_CALL_NAME, &j_name))
			fprintf(fp, "%s ", json_object_get_string(j_name));
		if (json_object_object_get_ex(j_entry, K_CALL_OFFSET, &j_offset))
			fprintf(fp, "+ %s ", json_object_get_string(j_offset));

		fprintf(fp,
		        "(%s) [%s] + %s\n",
		        json_object_get_string(j_addr),
		        json_object_get_string(j_m_name),
		        json_object_get_string(j_m_offset));
	}
	fprintf(fp, "End of Call Stack\n");
}

static void print_native_hash(json_object *j_root, FILE *fp)
{
	assert(j_root);
	assert(fp);

	json_object *j_native_hash;
	if (!json_object_object_get_ex(j_root, K_NATIVE_HASH, &j_native_hash)) {
		_W("No native hash in the JSON object");
		return;
	}

	json_object_get_string(j_native_hash);
	fprintf(fp,
		"STACK_SHA256: %s\n",
		json_object_get_string(j_native_hash));
}


void print_report_info(json_object *j_root, FILE *fp)
{
	assert(j_root);
	assert(fp);

	print_exe_path(j_root, fp);
	print_signal(j_root, fp);
	print_registers(j_root, fp);
	print_memory_info(j_root, fp);
	print_threads(j_root, fp);
	print_maps(j_root, fp);
	print_callstack(j_root, fp);
	print_native_hash(j_root, fp);
}

