/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Mateusz Moscicki <m.moscicki2@partner.samsung.com>
 */
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>

#define TMP_DIR "/tmp/"
#define DEFAULT_FILE_NAME "core_"

enum copy_errors {
	COPY_OK,
	COPY_WRITE_FAIL,
	COPY_READ_FAIL
};

static int copy_data(int dest, int src)
{
	char buff[1024];
	int readed;
	while ((readed = read(src, buff, sizeof(buff))) > 0) {
		int written = 0;
		while ((written = write(dest, buff, readed)) < readed) {
			if (written < 0)
				return COPY_WRITE_FAIL;
			readed -= written;
		}
	}

	return readed >= 0 ? COPY_OK : COPY_READ_FAIL;
}

int main(int argc, char *argv[])
{
#define EXIT_FAIL(fmt, ...) { \
		dprintf(STDERR_FILENO, fmt, ##__VA_ARGS__); \
		ret = EXIT_FAILURE; \
		goto exit; \
	}

	int ret = EXIT_SUCCESS;
	char *file_name;

	if (argc >= 2)
		file_name = strdup(argv[1]);
	else
		file_name = tempnam(TMP_DIR, DEFAULT_FILE_NAME);

	if (file_name == NULL)
		EXIT_FAIL("error: %m\n")

	int fd = open(file_name, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
	if (fd == -1)
		EXIT_FAIL("open file %s error: %m\n", file_name)

	int copy_res = copy_data(fd, STDIN_FILENO);
	close(fd);

	switch (copy_res) {
	case COPY_WRITE_FAIL:
		EXIT_FAIL("write to FD error: %m\n")
	case COPY_READ_FAIL:
		EXIT_FAIL("read from STDIN error: %m\n")
	}

	fd = open(file_name, O_RDONLY);
	if (fd == -1)
		EXIT_FAIL("open file %s error: %m\n", file_name)

	copy_res = copy_data(STDOUT_FILENO, fd);
	close(fd);

	switch (copy_res) {
	case COPY_WRITE_FAIL:
		EXIT_FAIL("write to STDOUT error: %m\n")
	case COPY_READ_FAIL:
		EXIT_FAIL("read from FD error: %m\n")
	}

exit:
	free(file_name);
	return ret;
#undef EXIT_FAIL
}

